;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; $Id: 10xselection.el,v 1.2 2004/01/17 14:26:03 lioka Exp $
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; workaround for CTEXT selection handling in XEmacs
; Author: Stephen J. Turnbull <stephen@xemacs.org>
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(MULElaunch
 (defconst sjt/extended-segment-parser-re
   (concat "\033%/"			; extended segment escape
	   "[0-4]"			; if we grok it, we should know width
	   "\\([\200-\377][\200-\377]\\)" ; length specification, base 128
	   "\\([^\002]*\\)"		; TODO: should match only Latin 1
	   "\002")			; ASCII STX = START OF TEXT
   "Regexp to detect and parse X Compound Text extended segment parameters.")
 
 (defconst sjt/coding-system-x11-aliases
   "Coding system aliases used in X Compound Text extended segments.

NB: we sanity check all uses of this mapping, so even if the particular
XEmacs doesn't know that coding system, please report as a bug if you see
a broken paste of the form \"^[%/1\200\227iso8859-15\002...\"."
   ;; TODO: add to this list.
  ;;;("koi8-r" . koi8-r)
   '(("iso8859-15" . iso-8859-15)
     ))
 
 (defun sjt/find-coding-system-from-alias (name)
   (setq name (downcase name))
   (cond ((find-coding-system (intern name)))
	 ((let ((pair (assoc name sjt/coding-system-x11-aliases)))
	    (when pair (find-coding-system (cdr pair)))))
	 ;; we could try guessing here, eg a prefix "iso" -> "iso-"
	 (t nil)))
 
 (defun sjt/select-convert-from-text (selection type value)
   "Converts a text selection to a Lisp string, handling X11 extended segments."
   (when (stringp value)
     (catch 'unconvertible
       (while (string-match sjt/extended-segment-parser-re value)
	 (let* ((cs (find-coding-system (sjt/find-coding-system-from-alias
					 (match-string 2 value))))
		(segstart (match-end 0))
		(lenspec (match-string 1 value))
		(length (+ (* 128 (- (char-to-int (aref lenspec 0)) 128))
			   (- (char-to-int (aref lenspec 1))
                             128
                             (length (match-string 2))
                             1)))
		(segend (+ segstart length)))
	   (unless cs (throw 'unconvertible nil))
	   (setq value (concat (substring value 0 (match-beginning 0))
			       (decode-coding-string
				(substring value segstart segend)
				cs)
			       ;; should eliminate while in favor of
			       ;; tail-recursion on this substring
			       (substring value segend))))))
     value))

 (defun sjt/toggle-compound-text-converter ()
   "Toggle the recognition and conversion of compound text extended segments."
   (interactive)
   (when (emacs-version>= 21 4)
     (let ((cvt (cond ((eq (cdr (assoc 'TEXT selection-converter-in-alist))
                           'select-convert-from-text)
		       'sjt/select-convert-from-text)
		      (t 'select-convert-from-text)))
	   (alist selection-converter-in-alist))
       ;; no mercy!!
       (while alist
	 (when (memq (cdar alist)
		     '(select-convert-from-text sjt/select-convert-from-text))
	   (setcdr (car alist) cvt))
        (setq alist (cdr alist)))
       (message "%s" cvt))))
 
 ;; OK, do it!
 (sjt/toggle-compound-text-converter)
)
